"""
Author Aurore Guillevic
email  aurore.guillevic@inria.fr
date   March 16, 2019

Example of NFS with Joux-Lercier polynomial selection for p=1109.
"""
# import printing functions for latex output
from examples_pretty_print import pretty_print_product, pretty_print_tex_product, pretty_print_units

p = 1109     # prime, p = 1 mod 4
r = (p-1)//4 # prime order of the subgroup
Fr = GF(r)   # for the linear algebra
Fp = GF(p)
g_ = Fp(2)   # a generator of Fp*
g4 = g_^((p-1)//4) # g4 is of order 4 in Fp*

ZZx.<x> = ZZ[]

print("p = {} is prime".format(p))
print("(p-1)/4 = {} is prime".format(r))

print("Polynomial selection of Joux--Lercier")

tab_h_3 = [      # a set of irreducible polynomials of degree 3 and small coefficients
    x**3 + x + 1,
    x**3 - x + 1,
    x**3 + x**2 + x - 1]

Fpz.<z> = Fp[]
for f in tab_h_3:
    fp = Fpz(f)
    print("f = {} = {} mod p".format(f, fp.factor()))
    Kf.<af> = NumberField(f)
    Clf = Kf.class_number()
    print("Cl(Kf) = {}".format(Clf))
    print("units(Kf) = {}".format(Kf.units()))
    for m,_ in fp.roots():
        m0 = ZZ(p-m)
        M = Matrix(ZZ, 3,3, [p,0,0, m0,1,0, 0,m0,1])
        M = M.LLL()
        print("m = {}\nM = \n{}".format(m,M))
        id_r=0
        for ri in M.rows():
            ri0 = ri.list()
            gr = gcd(ri0)
            if gr > 1:
                ri0 = [rj//gr for rj in ri0]
            g = ZZx(ri0)
            if g.degree() == 2 and g.is_irreducible():
                print("\nrow {}: g={}".format(id_r,g))
                Kg.<ag> = NumberField(g)
                Clg = Kg.class_number()
                print("  disc(g) = {} = {}".format(g.disc(), g.disc().factor()))
                print("  Cl(Kg) = {}".format(Clg))
                print("  units(Kg) = {}".format(Kg.units()))
                print("  real, cpx roots {}".format(Kg.signature()))
            id_r += 1
    print("")

# you may try these polynomial pairs:
#f = x^3+x+1 ; g = x^2 + 9*x - 3
#f = x^3-x+1 ; g = 6 + 6*x + 5*x^2
f = x^3-x+1 ; g = -11 + 5*x + x^2
print("chosen: f = {}, g = {}".format(f,g))

mp = gcd(Fpz(f), Fpz(g)).roots()[0][0]
m0 = ZZ(mp) # common root
print("res(f,g) = {}".format(f.resultant(g)))
disc_f =  f.discriminant()
print("disc(f) = {}".format(disc_f))
disc_g =  g.discriminant()
print("disc(g) = {}".format(disc_g))
lc_g = g.leading_coefficient()

Bf=23
Bg=23

Ff = dict()
Fg = dict()
print("factor basis f")
for l in prime_range(2, Bf+1):
    Fl = GF(l)
    Flz.<z> = Fl[]
    fl = Flz(f)
    if len(fl.roots())>0:
        Ff[l] = [ZZx(fi) for (fi,ei) in fl.factor() if fi.degree() == 1]
    print("{}: {}".format(l, (fl).factor()))

print("\nfactor basis g")
for l in prime_range(2, Bg+1):
    Fl = GF(l)
    Flz.<z> = Fl[]
    gl = Flz(g)
    if len(gl.roots())>0:
        Fg[l] = [ZZx(gi) for (gi,ei) in gl.factor() if gi.degree() == 1]
    print("{}: {}".format(l, (gl).factor()))
print("")

A = 11 # a,b in the relation collection step: a in [-A,A], b in [1,A]

print Ff
print Fg
print("chosen smoothness bounds: f-side {}, g-side {}".format(Bf,Bg))

Kf.<af> = NumberField(f)
Kg.<ag> = NumberField(g)
Clf = Kf.class_number()
Clg = Kg.class_number()
print("Class number(Kf) = {}".format(Clf))
print("Class number(Kg) = {}".format(Clg))
Of = Kf.maximal_order()
Og = Kg.maximal_order()
"""
F_f = dict()
for (l,li) in Ff.items():
    for d in li:
        Il = Kf.ideal([l,d])
        gen_Il = Il.gens_reduced()
        if len(gen_Il) == 1:
            if F_f.has_key(l):
                F_f[l].append(gen_Il[0])
            else:
                F_f[l] = [gen_Il[0]]
"""
# Problem: Sage chooses generators of the form af^2+af+1
F_f = {5:[af+2], 7:[af-2], 11:[2*af+1], 17:[3*af-2], 19:[3*af-1], 23:[af+3,2*af-3]}
                
F_g = dict()
for (l,li) in Fg.items():
    for d in li:
        Il = Kg.ideal([l,d])
        gen_Il = Il.gens_reduced()
        if len(gen_Il) == 1:
            if F_g.has_key(l):
                F_g[l].append(gen_Il[0])
            else:
                F_g[l] = [gen_Il[0]]

fond_unit_f = Kf.units()[0]
fond_unit_g = Kg.units()[0]

idx = 0 # is for the unit
id_unit_f = idx
idx_ideal_f = dict()
idx_ideal_f[(1,fond_unit_f.polynomial())] = 0
idx = 1
for (l,li) in Ff.items():
    for id_li in li:
        idx_ideal_f[(l,id_li)] = idx
        idx += 1

idx_ideal_g = dict()
for (l,li) in Fg.items():
    for id_li in li:
        idx_ideal_g[(l,id_li)] = idx
        idx += 1
idx_ideal_g[(1, fond_unit_g.polynomial())] = idx
id_unit_g = idx
idx += 1

print idx_ideal_f
print idx_ideal_g

# http://doc.sagemath.org/html/en/reference/number_fields/sage/rings/number_field/unit_group.html
print "\nUnit group of Kf"
UKf = Kf.unit_group()
u0_f,u1_f = UKf.gens()
torsion_unit_f = Kf(u0_f)
fond_unit_f = Kf(u1_f)
print("the generators of the unit group are ({},{}), in Kf: ({},{})".format(u0_f, u1_f, torsion_unit_f, fond_unit_f))
print("the generators of the unit group are of order: {}".format([ui_f.multiplicative_order() for ui_f in UKf.gens()]))
print("the unit group has rank {}".format(UKf.rank()))

# http://doc.sagemath.org/html/en/reference/number_fields/sage/rings/number_field/unit_group.html
print "\nUnit group of Kg"
UKg = Kg.unit_group()
u0_g,u1_g = UKg.gens()
torsion_unit_g = Kg(u0_g)
fond_unit_g = Kg(u1_g)
print("the generators of the unit group are ({},{}), in Kg: ({},{})".format(u0_g, u1_g, torsion_unit_g, fond_unit_g))
print("the generators of the unit group are of order: {}".format([ui_g.multiplicative_order() for ui_g in UKg.gens()]))
print("the unit group has rank {}".format(UKg.rank()))

row_id = 0
rows = idx
cols = idx
print("{} columns, {} relations meeded".format(cols,rows))
if lc_g == 1:
    mat = [[0 for j in range(cols)] for i in range(rows)]
else:
    mat = [[0 for j in range(cols-1)]+[1] for i in range(rows)]
    
print("start relation collection, a in [-{0},{0}], b in [1,{0}]".format(A))
for a in range(-A,A+1):
    for b in [b for b in range(1,A+1) if gcd(a,b) == 1]:
        nf = ZZ(f.resultant(a-b*x))
        ng = ZZ(g.resultant(a-b*x))
        if nf == 0 or ng == 0:
            continue
        fac_f = nf.factor()
        fac_g = ng.factor()
        j = 0
        while j < len(fac_f) and fac_f[j][0] <= Bf:
            j += 1
        smooth_f = (j >= len(fac_f))
        j = 0
        while j < len(fac_g) and fac_g[j][0] <= Bg:
            j += 1
        smooth_g = (j >= len(fac_g))
        if smooth_f and smooth_g:
            #print("nf = {} ng = {}".format(nf,ng))
            #print("(a,b) = ({},{})".format(a,b))
            # compute the factorization into prime ideals
            if row_id >= rows: # add a row
                if lc_g == 1:
                    mat.append([0 for j in range(cols)])
                else:
                    mat.append([0 for j in range(cols-1)]+[1])
            str_f = ""
            # prod_id_f will be the product of generators of ideals,
            # so that a-b*af = prod_id_f up to a unit to compute: u = (a-b*af) / prod_id_f
            prod_id_f = Kf(1)
            for (l,ei) in fac_f: # l a prime, ei an exponent: l^ei | fac_f
                # to which ideal above l does the factor (l^ei) correspond to?
                Fl = GF(l)
                Flz.<z> = Fl[]
                ideal_f = gcd(Flz(f), Flz(a-b*x))
                str_f += "("+str(l)+","+str(ideal_f)+")"
                if ei>1: str_f+= "^"+str(ei)
                for id_x in F_f[l]: # identify which ones of the ideals of norm l are actually involved in a-b*x
                    if gcd(Flz(id_x.polynomial()), ideal_f) > 1:
                        prod_id_f *= id_x^ei
                        mat[row_id][idx_ideal_f[(l,ZZx(ideal_f))]] = ei
            unf = (a-b*af)/prod_id_f # unit
            if not abs(unf.norm()) == 1:
                print("error in computing (a-b*af)/(prod(generator ideals))")
            # compute the contribution of the units: u0=-1 and the fundamental unit u1
            (log_u0, log_u1) = UKf.log(unf)
            # this should be: (-1)^log_u0*(uf)^log_u1 == unf
            if log_u0 == 0: unfs=" "
            else: unfs="-"            
            mat[row_id][id_unit_f] = log_u1 # exponent
            str_uf = pretty_print_units(log_u1, unfs, "f")

            str_g = ""
            prod_id_g = Kg(1)
            for (l,ei) in fac_g:
                Fl = GF(l)
                Flz.<z> = Fl[]
                ideal_g = gcd(Flz(g), Flz(a-b*x))
                str_g += "("+str(l)+","+str(ideal_g)+")"
                if ei>1: str_g+= "^"+str(ei)
                for id_x in F_g[l]:
                    if gcd(Flz(id_x.polynomial()), ideal_g) > 1:
                        prod_id_g *= id_x^ei
                        mat[row_id][idx_ideal_g[(l,ZZx(ideal_g))]] = -ei
                #if lc_g > 1 and (lc_g % l == 0)  and ideal_g == 1:
                #    mat[row_id][idx_ideal_g[(l,0)]] -= ei
                #print("({},{})".format(l,ei))
            ung = (a-b*ag)/prod_id_g # unit
            
            if not abs(ung.norm()) == 1:
                print("error in computing (a-b*ag)/(prod(generator ideals))")
            # compute the contribution of the units: u0=-1 and the fundamental unit u1
            (log_u0, log_u1) = UKg.log(ung)
            # this should be: (-1)^log_u0*(ug)^log_u1 == ung
            if log_u0 == 0: ungs=" "
            else: ungs="-"            
            mat[row_id][id_unit_g] = -log_u1 # exponent
            str_ug = pretty_print_units(log_u1, ungs, "g")

            print("{:3},{:2}&{:5}& {:12s}&{:21s}&{}&  {:5}&{:15s}&{:29s}&{} \\\\".format(a,b,nf,pretty_print_tex_product(nf,fac_f), str_f, str_uf, ng,pretty_print_tex_product(ng,fac_g), str_g, str_ug))
            row_id += 1

print("{} relations found".format(row_id))
if row_id >= cols:
    M = Matrix(ZZ, row_id, cols, mat)
    print M
    latex.matrix_delimiters(left='[', right=']')
    print latex(M) # see http://doc.sagemath.org/html/en/tutorial/latex.html
    Mr = Matrix(Fr, row_id, cols, mat)
    print("M mod r has rank {}".format(Mr.rank()))
    v = Mr.right_kernel()
    print v
    v = v[1]
"""
22 relations found
[ -9   0   2   0   0   1   0   0   1   0   0   1   0   0   1   0   0   0   0]
[-13   0   0   0   2   0   0   0   0   0   1   1   0   0   0   0   0   1  -1]
[ -2   0   0   1   0   1   0   0   0   1   0   0   0   0   0   0   0   0   0]
[ -6   0   1   0   1   0   0   0   0   0   0   0   1   0   0   0   0   0   0]
[-13   0   0   0   0   0   0   0   0   0   0   0   1   0   1   0   0   0   0]
[  3   2   0   0   0   1   0   0   1   0   0   0   0   1   0   0   1   0   1]
[  0   0   0   0   0   0   1   0   0   0   0   0   0   0   0   0   1   0   0]
[ -8   0   1   0   0   0   0   0   0   0   1   0   0   1   0   0   0   0   0]
[  0   1   0   0   0   0   0   0   0   0   0   0   0   0   0   1   0   0   0]
[ -4   0   0   0   0   0   0   0   1   1   0   0   0   0   0   0   0   0   1]
[  1   0   0   0   0   0   0   0   0   0   0   1   0   0   0   0   0   0   0]
[  3   0   0   0   0   0   0   0   0   0   1   0   0   0   0   0   0   0   0]
[  6   1   0   0   0   0   0   0   1   0   0   0   1   0   0   0   0   0   1]
[  0   0   1   0   0   0   0   0   1   0   0   0   0   0   0   0   0   0   1]
[  0   0   0   0   1   0   0   0   0   1   0   0   0   1   0   0   0   0   1]
[  8   2   0   0   0   0   0   0   0   0   0   0   0   0   1   0   0   0   0]
[  0   0   0   0   0   0   0   1   0   1   0   0   0   0   0   0   0   0   1]
[  7   1   0   1   0   0   0   0   0   0   0   0   0   0   0   0   0   1   0]
[  3   0   0   2   0   0   0   0   1   0   0   0   0   1   0   0   0   0   1]
[  3   0   1   0   0   0   1   0   0   0   0   0   0   0   1   1   0   0   0]
[ -4   0   0   0   1   1   0   0   1   0   2   0   0   0   0   0   0   0   0]
[  2   0   0   0   0   1   1   0   0   0   0   0   1   0   0   0   0   0   1]
"""    
    # [  1  21  90  83 130 102 255  51  55  94 274 276  63 198 227 256  22 166 132]
    #    uf  5   7  11  17  19  23  23 | 3   5   5  11  11  13  13  17  17  23  ug

g_ = 2
t = 314
deg_f = f.degree()
deg_g = g.degree()

for ei in [1,12]:
    gi = ZZ(Fp(g_)^ei)
    MDg2 = Matrix(ZZ, 4, 4, [p  ,0 , 0,0, \
                             -m0,1 , 0,0, \
                             gi ,0 , 1,0, \
                             0  ,gi, 0,1])

    print("descent of generator g={} = 2^{}".format(gi,ei))
    print("M = \n{}".format(MDg2))
    Mg2R = MDg2.LLL()
    print Mg2R
    ri_n = 0
    for ri in Mg2R.rows():
        ri_n += 1
        rri = ri.list()
        if rri != [2,0,1,0] and rri != [0,2,0,1]:
            num = ZZx(rri[:2])
            den = ZZx(rri[2:])
            if Fp(num(m0)/den(m0)) == Fp(gi):
                print("row {} {}".format(ri_n, rri))
                print "  num = ", num, "norm =", ZZ(g.resultant(num)).factor()
                print "  den = ", den, "norm =", ZZ(g.resultant(den)).factor()
                print("  ({})/({})(m0) = {}, gi = {}".format(num,den, Fp(num(m0)/den(m0)), Fp(gi)))
            else:
                print("  pb ({})/({})(m0) = {} != {}".format(num,den,m0, Fp(num(m0)/den(m0)), Fp(gi)))
"""
descent of generator g=769 = 2^12
[-3  0 -2  1]
[ 1  1 -3  2]
[-1  0 -3 -5]
[-1 14  2 -2]
row 2 [1, 1, -3, 2]
  num =  x + 1 norm = -1 * 3 * 5
  den =  2*x - 3 norm = -1 * 5
  (x + 1)/(2*x - 3)(m0) = 769, gi = 769
"""
(-1-ag) == (ag+7)*(ag+6)*(3*ag-5)
(3-2*ag) == (ag+6)*(3*ag-5)

log_g12 = v[idx_ideal_g[(3,x+1)]]
log_g0 = Fr(log_g12/12)
print("log g0 = {}/12 = {} mod r".format(log_g12, log_g0))

print("descent of target t={}".format(t))
MDt2 = Matrix(ZZ, 4, 4, [p  ,0, 0,0, \
                         -m0,1, 0,0, \
                         t  ,0, 1,0, \
                         0  ,t, 0,1])
Mt2R = MDt2.LLL()
print Mt2R
# let's take -x/(2-3*x)
Fp(-m0/(2-3*m0)) == Fp(t)
(2-3*ag) == (ag+6)*(ag+8)*(3*ag-5)

# v = ( 1,  21,  90,   83,  130,  102, 255,   51,    222, 183,   3, 1, 214,  79,  50,  21, 255,   111, 145)
#      af,af+2,af-2,2af+1,3af-2,3af-1,af+3,2af-3,   ag+7,ag+6,ag-1,ag,ag+5,ag+8,ag-3,ag+2,ag+3,-3ag+4,3ag-5

log_t = v[idx_ideal_g[(11,x)]] - v[idx_ideal_g[(5,x+1)]] - v[idx_ideal_g[(13,x+8)]] - v[id_unit_g]
log_t = log_t % r
print("log_t = {} - {} - {} - {} mod {} = {}".format(v[idx_ideal_g[(11,x)]], v[idx_ideal_g[(5,x+1)]], v[idx_ideal_g[(13,x+8)]], v[id_unit_g], r, log_t))
print("log_t = {} mod {}".format(log_t, r))

log_g0_t = ZZ(Fr(log_t/log_g0))
print("log_g0(t) = log_t/log_g0  = {}/{} = {} mod r".format(log_t, log_g0, log_g0_t))

frac = Fp(t)/Fp(2)^log_g0_t
print("t/2^log_2(t) = {}".format(frac))
if frac == g4:
    log_g0_t += r
    corr = r
elif frac == g4^2:
    log_g0_t += 2*r
    corr = 2*r
elif frac == g4^3:
    log_g0_t += 3*r
    corr = 3*r
log_g0_t = log_g0_t % (p-1)
print("log_g0(t) <- log_g0_t + {} = {} mod (p-1)".format(corr, log_g0_t))
frac = Fp(t)/Fp(2)^log_g0_t
print("t/2^log_2(t) = {}".format(frac))


# a small example of descent with very simple lift

p = 1109
t = 314
m = -347
t+p == 35 - 4*m

g0 = 2
Bg = 23

bad = 0
tt = t
i=1
is_smooth=False
while not is_smooth:
    tt = (g0*tt) % p
    if abs(tt) < abs(m) :
        M = Matrix(ZZ, 2,2, [tt+p,0,-m,1])
    else:
        M = Matrix(ZZ, 2,2, [tt,0,-m,1])
    R = M.LLL()
    u = R[0][0]
    v = R[0][1]
    if ((u + v*m) % p) == tt:
        Ng = ZZ((u+v*ag).norm())
        fact_Ng = Ng.factor()
        j = 0
        while j < len(fact_Ng) and fact_Ng[j][0] <= Bg:
            j += 1
        is_smooth = (j >= len(fact_Ng))
        print("g0^{:2d}*t = {:4d}, Norm({:3d}{:+3d}*ag) = {} = {}".format(i,tt,u,v,Ng,fact_Ng))
    else:
        bad += 1
    i += 1

bad = 0
tt = 1
i=1
is_smooth=False
while not is_smooth:
    tt = (g0*tt) % p
    if abs(tt) < abs(m) :
        M = Matrix(ZZ, 2,2, [tt+p,0,-m,1])
    else:
        M = Matrix(ZZ, 2,2, [tt,0,-m,1])
    R = M.LLL()
    u = R[0][0]
    v = R[0][1]
    if ((u + v*m) % p) == tt:
        Ng = ZZ((u+v*ag).norm())
        fact_Ng = Ng.factor()
        j = 0
        while j < len(fact_Ng) and fact_Ng[j][0] <= Bg:
            j += 1
        is_smooth = (j >= len(fact_Ng))
        print("g0^{:2d}*t = {:4d}, Norm({:3d}{:+3d}*ag) = {} = {}".format(i,tt,u,v,Ng,fact_Ng))
    else:
        bad += 1
    i += 1

(-17-3*ag)/(ag-1)*(ag-3)

#descent on f-side

p = 1109
t = 314
m = -347
g0 = 2
f = x^3-x+1

M = Matrix(ZZ, 6,6, [p,0,0,0,0,0, -m,1,0,0,0,0,  0,-m,1,0,0,0,  t,0,0,1,0,0, 0,t,0,0,1,0, 0,0,t,0,0,1])
R = M.LLL()
R

"""
[-1  0  0 -1  0 -2]
[-1  0  1 -1 -2  1]
[ 0 -1  0  2 -3  0]
[ 3 -1  1  0 -1 -1]
[-2 -1  3  1  1  0]
[-2 -4 -1  1  1 -1]
"""
Fp = GF(p)
Fp((-1+m**2)/(-1-2*m+m**2))
Fp((3-m+m**2)/(-m-m**2))
Fp((-2-m+3*m**2)/(1+m))
Fp((-2-4*m-m**2)/(1+m-m^2))

(-1+x^2).factor()     # (x-1)(x+1)
(-1-2*x+x^2).factor() # irr
(3-x+x^2).factor()    # irr
(-x-x^2).factor()     # -x*(x-1)
(-2-x+3*x^2).factor() # (x-1)*(3*x+2)
(-2-4*x-x^2).factor() # irr
(1+x-x^2).factor()    # irr

Kf.<af> = NumberField(f)
((-2-4*x-x^2)(af)).norm().factor() # 7^2
((1+x-x^2)(af)).norm().factor() # -5

(-af^2 - 4*af - 2) == -(af-2)^2*af^(-7)
(-af^2 + af + 1) == -(af+2)*af^4


gi = g0^2
M = Matrix(ZZ, 6,6, [p,0,0,0,0,0, -m,1,0,0,0,0,  0,-m,1,0,0,0,  gi,0,0,1,0,0, 0,gi,0,0,1,0, 0,0,gi,0,0,1])
R = M.LLL()
R
"""
with gi = g0:
[ 2  0  0  1  0  0]
[ 0  2  0  0  1  0]
[ 0  0  2  0  0  1]
[ 2  2  1 -2 -2 -2]
[ 1  2 -1  0 -4  2]
[-3  1  1  4 -2  0]

with gi = g0^2:
[-1  0 -2  1 -1  2]
[-1 -2  1  0  2 -1]
[-1 -1 -1 -3  1  0]
[ 2 -2 -1  2  1  1]
[-2  2 -1  1  2  1]
[-1  0  2  1 -1  3]

"""
for ri in range(6):
    ax = ZZx(list(R[ri][:3]))
    bx = ZZx(list(R[ri][3:]))
    #if ax.is_irreducible() and bx.is_irreducible() :
    print("row {}".format(ri))
    aa = ax(af)
    bb = bx(af)
    print("a = {:18s}, norm {} = {}, b = {:18s}, norm {}={}".format(aa,aa.norm(), aa.norm().factor(), bb, bb.norm(), bb.norm().factor()))

# row 4
# a = -af^2 + 2*af - 2  , norm -5  = -1 * 5 , b = af^2 + 2*af + 1   , norm 1  = 1
a = -af^2 + 2*af - 2
a5 = af+2 # of norm 5
b = af^2 + 2*af + 1
# this is a power of a unit

a/a5 == -af^8
a == -a5*af^8
b == af^(-8)

log_g0 = 21 + 8 + 8
